from .account import Account
from .neato import Neato
from .robot import Robot
from .session import OAuthSession, PasswordlessSession, PasswordSession
from .version import __version__
from .vorwerk import Vorwerk



"""Account access and data handling for beehive endpoint."""

import logging
import os
import shutil

import requests
from voluptuous import (
    ALLOW_EXTRA,
    All,
    Any,
    Extra,
    MultipleInvalid,
    Optional,
    Range,
    Required,
    Schema,
    Url,
)

from .exceptions import NeatoRobotException, NeatoUnsupportedDevice
from .robot import Robot
from .session import Session

_LOGGER = logging.getLogger(__name__)

ROBOT_SCHEMA = Schema(
    {
        Required("serial"): str,
        "prefix": Any(str, None),
        Required("name"): str,
        "model": Any(str, None),
        Required("secret_key"): str,
        "purchased_at": Any(str, None),
        "linked_at": Any(str, None),
        Required("traits"): list,
        # Everything below this line is not documented, but still present
        "firmware": Any(str, None),
        "timezone": Any(str, None),
        Required("nucleo_url"): Url,
        "mac_address": Any(str, None),
        "created_at": Any(str, None),
    },
    extra=ALLOW_EXTRA,
)
MAP_SCHEMA = Schema(
    {
        "version": Any(int, None),
        Required("id"): str,
        Required("url"): Url,
        "url_valid_for_seconds": Any(int, None),
        Optional("run_id"): str,  # documented, but not present
        "status": Any(str, None),
        "launched_from": Any(str, None),
        "error": Any(str, None),
        "category": Any(int, None),
        "mode": Any(int, None),
        "modifier": Any(int, None),
        "start_at": Any(str, None),
        "end_at": Any(str, None),
        "end_orientation_relative_degrees": All(int, Range(min=0, max=360)),
        "run_charge_at_start": All(int, Range(min=0, max=100)),
        "run_charge_at_end": All(int, Range(min=0, max=100)),
        "suspended_cleaning_charging_count": Any(int, None),
        "time_in_suspended_cleaning": Any(int, None),
        "time_in_error": Any(int, None),
        "time_in_pause": Any(int, None),
        "cleaned_area": Any(float, None),
        "base_count": Any(int, None),
        "is_docked": Any(bool, None),
        "delocalized": Any(bool, None),
        # Everything below this line is not documented, but still present
        "generated_at": Any(str, None),
        "persistent_map_id": Any(str, None),
        "cleaned_with_persistent_map_id": Any(str, None),
        "valid_as_persistent_map": Any(bool, None),
        "navigation_mode": Any(int, None),
    },
    extra=ALLOW_EXTRA,
)
MAPS_SCHEMA = Schema(
    {"stats": {Extra: object}, Required("maps"): [MAP_SCHEMA]},
    extra=ALLOW_EXTRA,
)
PERSISTENT_MAP_SCHEMA = Schema(
    {
        Required("id"): str,
        Required("name"): str,
        Required("url"): Url,
        "raw_floor_map_url": Any(Url, None),
        "url_valid_for_seconds": Any(int, None),
    },
    extra=ALLOW_EXTRA,
)
PERSISTENT_MAPS_SCHEMA = Schema(Required([PERSISTENT_MAP_SCHEMA]))


class Account:
    """
    Class with data and methods for interacting with a pybotvac cloud session.

    :param email: Email for pybotvac account
    :param password: Password for pybotvac account

    """

    def __init__(self, session: Session):
        """Initialize the account data."""
        self._robots = set()
        self.robot_serials = {}
        self._maps = {}
        self._persistent_maps = {}
        self._session = session

    @property
    def robots(self):
        """
        Return set of robots for logged in account.

        :return:
        """
        if not self._robots:
            self.refresh_robots()

        return self._robots

    @property
    def maps(self):
        """
        Return set of userdata for logged in account.

        :return:
        """
        self.refresh_maps()

        return self._maps

    def refresh_maps(self):
        """
        Get information about maps of the robots.

        :return:
        """

        for robot in self.robots:
            url = f"users/me/robots/{robot.serial}/maps"
            resp2 = self._session.get(url)
            try:
                MAPS_SCHEMA(resp2.json())
                self._maps.update({robot.serial: resp2.json()})
            except MultipleInvalid as ex:
                _LOGGER.warning(
                    "Invalid response from %s: %s. Got: %s", url, ex, resp2.json()
                )

    def refresh_robots(self):
        """
        Get information about robots connected to account.

        :return:
        """

        resp = self._session.get("users/me/robots")

        for robot in resp.json():
            _LOGGER.debug("Create Robot: %s", robot)
            try:
                ROBOT_SCHEMA(robot)
                robot_object = Robot(
                    name=robot["name"],
                    vendor=self._session.vendor,
                    serial=robot["serial"],
                    secret=robot["secret_key"],
                    traits=robot["traits"],
                    endpoint=robot["nucleo_url"],
                )
                self._robots.add(robot_object)
            except MultipleInvalid as ex:
                # Robot was not described accordingly by neato
                _LOGGER.warning(
                    "Bad response from robots endpoint: %s. Got: %s", ex, robot
                )
                continue
            except NeatoUnsupportedDevice:
                # Robot does not support home_cleaning service
                _LOGGER.warning("Your robot %s is unsupported.", robot["name"])
                continue
            except NeatoRobotException:
                # The state of the robot could not be received
                _LOGGER.warning("Your robot %s is offline.", robot["name"])
                continue

        self.refresh_persistent_maps()
        for robot in self._robots:
            robot.has_persistent_maps = robot.serial in self._persistent_maps

    @staticmethod
    def get_map_image(url, dest_path=None, file_name=None):
        """
        Return a requested map from a robot.

        :return:
        """

        try:
            image = requests.get(url, stream=True, timeout=10)

            if dest_path:
                image_url = url.rsplit("/", 2)[1] + "-" + url.rsplit("/", 1)[1]
                if file_name:
                    image_filename = file_name
                else:
                    image_filename = image_url.split("?")[0]

                dest = os.path.join(dest_path, image_filename)
                image.raise_for_status()
                with open(dest, "wb") as data:
                    image.raw.decode_content = True
                    shutil.copyfileobj(image.raw, data)
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            raise NeatoRobotException("Unable to get robot map") from ex

        return image.raw

    @property
    def persistent_maps(self):
        """
        Return set of persistent maps for logged in account.

        :return:
        """
        self.refresh_persistent_maps()

        return self._persistent_maps

    def refresh_persistent_maps(self):
        """
        Get information about persistent maps of the robots.

        :return:
        """

        for robot in self._robots:
            url = f"users/me/robots/{robot.serial}/persistent_maps"
            resp2 = self._session.get(url)

            try:
                PERSISTENT_MAPS_SCHEMA(resp2.json())
                self._persistent_maps.update({robot.serial: resp2.json()})
            except MultipleInvalid as ex:
                _LOGGER.warning(
                    "Invalid response from %s: %s. Got: %s", url, ex, resp2.json()
                )



import hashlib
import hmac
import logging
import re
from datetime import datetime, timezone
from email.utils import format_datetime

import requests
import urllib3
from voluptuous import (
    ALLOW_EXTRA,
    All,
    Any,
    Extra,
    MultipleInvalid,
    Range,
    Required,
    Schema,
)

from .exceptions import NeatoRobotException, NeatoUnsupportedDevice
from .neato import Neato  # For default Vendor argument

# Disable warning due to SubjectAltNameWarning in certificate
urllib3.disable_warnings(urllib3.exceptions.SubjectAltNameWarning)

_LOGGER = logging.getLogger(__name__)

SUPPORTED_SERVICES = ["basic-1", "minimal-2", "basic-2", "basic-3", "basic-4"]
ALERTS_FLOORPLAN = [
    "nav_floorplan_load_fail",
    "nav_floorplan_localization_fail",
    "nav_floorplan_not_created",
]

RESULT_SCHEMA = Schema(
    Any(
        "ok",
        "invalid_json",
        "bad_request",
        "command_not_found",
        "command_rejected",
        "ko",
        # Everything below this line is not documented, but still present
        "not_on_charge_base",
    )
)
STANDARD_SCHEMA = Schema(
    {
        "version": int,
        "reqId": str,
        Required("result"): RESULT_SCHEMA,
        "data": {Extra: object},
    },
    extra=ALLOW_EXTRA,
)
STATE_SCHEMA = Schema(
    {
        "version": int,
        "reqId": str,
        Required("result"): RESULT_SCHEMA,
        "data": {Extra: object},
        Required("state"): int,
        "action": int,
        "error": Any(str, None),
        "alert": Any(str, None),
        "cleaning": {
            "category": int,
            "mode": int,
            "modifier": int,
            "navigationMode": int,
            "spotWidth": int,
            "spotHeight": int,
        },
        "details": {
            "isCharging": bool,
            "isDocked": bool,
            "dockHasBeenSeen": bool,
            "charge": All(int, Range(min=0, max=100)),
            "isScheduleEnabled": bool,
        },
        "availableCommands": {
            "start": bool,
            "stop": bool,
            "pause": bool,
            "resume": bool,
            "goToBase": bool,
        },
        Required("availableServices"): {
            "findMe": str,
            "generalInfo": str,
            "houseCleaning": str,
            "localStats": str,
            "manualCleaning": str,
            "maps": str,
            "preferences": str,
            "schedule": str,
            "spotCleaning": str,
            # Undocumented services
            "IECTest": str,
            "logCopy": str,
            "softwareUpdate": str,
            "wifi": str,
        },
        "meta": {"modelName": str, "firmware": str},
    },
    extra=ALLOW_EXTRA,
)


class Robot:
    """Data and methods for interacting with a Neato Botvac Connected vacuum robot"""

    def __init__(
        self,
        serial,
        secret,
        traits,
        vendor=Neato,
        name="",
        endpoint="https://nucleo.neatocloud.com:4443",
        has_persistent_maps=False,
    ):
        """
        Initialize robot

        :param serial: Robot serial
        :param secret: Robot secret
        :param name: Name of robot (optional)
        :param traits: Extras the robot supports
        """
        self.name = name
        self._vendor = vendor
        self.serial = serial
        self.secret = secret
        self.traits = traits
        self.has_persistent_maps = has_persistent_maps

        # pylint: disable=anomalous-backslash-in-string
        self._url = "{endpoint}/vendors/{vendor_name}/robots/{serial}/messages".format(
            endpoint=re.sub(":\d+", "", endpoint),  # noqa: W605, Remove port number
            vendor_name=vendor.name,
            serial=self.serial,
        )
        self._headers = {"Accept": vendor.nucleo_version}

        # Check if service_version is supported
        # We manually scan the state here to perform appropriate error handling
        state = self.get_robot_state().json()
        if (
            "availableServices" not in state
            or "houseCleaning" not in state["availableServices"]
            or state["availableServices"]["houseCleaning"] not in SUPPORTED_SERVICES
        ):
            raise NeatoUnsupportedDevice(
                "Service houseCleaning is not supported by your robot"
            )

    def __str__(self):
        return "Name: %s, Serial: %s, Secret: %s Traits: %s" % (
            self.name,
            self.serial,
            self.secret,
            self.traits,
        )

    def _message(self, json: dict, schema: Schema):
        """
        Sends message to robot with data from parameter 'json'
        :param json: dict containing data to send
        :return: server response
        """

        try:
            response = requests.post(
                self._url,
                json=json,
                verify=self._vendor.cert_path,
                auth=Auth(self.serial, self.secret),
                headers=self._headers,
            )
            response.raise_for_status()
            schema(response.json())
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as ex:
            raise NeatoRobotException("Unable to communicate with robot") from ex
        except MultipleInvalid as ex:
            _LOGGER.warning(
                "Invalid response from %s: %s. Got: %s", self._url, ex, response.json()
            )

        return response

    def start_cleaning(
        self, mode=2, navigation_mode=1, category=None, boundary_id=None, map_id=None
    ):
        # mode & navigation_mode used if applicable to service version
        # mode: 1 eco, 2 turbo
        # navigation_mode: 1 normal, 2 extra care, 3 deep
        # category: 2 non-persistent map, 4 persistent map
        # boundary_id: the id of the zone to clean
        # map_id: the id of the map to clean

        # Default to using the persistent map if we support basic-3 or basic-4.
        if category is None:
            category = (
                4
                if self.service_version in ["basic-3", "basic-4"]
                and self.has_persistent_maps
                else 2
            )

        if self.service_version == "basic-1":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {"category": category, "mode": mode, "modifier": 1},
            }
        elif self.service_version == "basic-3" or "basic-4":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {
                    "category": category,
                    "mode": mode,
                    "modifier": 1,
                    "navigationMode": navigation_mode,
                },
            }
            if boundary_id:
                json["params"]["boundaryId"] = boundary_id
            if map_id:
                json["params"]["mapId"] = map_id
        elif self.service_version == "minimal-2":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {"category": category, "navigationMode": navigation_mode},
            }
        else:  # self.service_version == 'basic-2'
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {
                    "category": category,
                    "mode": mode,
                    "modifier": 1,
                    "navigationMode": navigation_mode,
                },
            }

        response = self._message(json, STATE_SCHEMA)
        result = response.json().get("result", None)
        alert = response.json().get("alert", None)
        if result != "ok":
            _LOGGER.warning(
                "Result of robot.start_cleaning is not ok: %s, alert: %s", result, alert
            )

        # Fall back to category 2 if we tried and failed with category 4
        if (
            category == 4
            and alert in ALERTS_FLOORPLAN
            or result == "not_on_charge_base"
        ):
            json["params"]["category"] = 2
            response_fallback = self._message(json, STATE_SCHEMA)
            result = response_fallback.json().get("result", None)
            alert = response_fallback.json().get("alert", None)
            if result != "ok":
                _LOGGER.warning(
                    "Result of robot.start_cleaning is not ok after fallback: %s, alert: %s",
                    result,
                    alert,
                )
            return response_fallback

        return response

    def start_spot_cleaning(self, spot_width=400, spot_height=400, mode=2, modifier=2):
        # Spot cleaning if applicable to version
        # spot_width: spot width in cm
        # spot_height: spot height in cm

        if self.spot_cleaning_version == "basic-1":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {
                    "category": 3,
                    "mode": mode,
                    "modifier": modifier,
                    "spotWidth": spot_width,
                    "spotHeight": spot_height,
                },
            }
        elif self.spot_cleaning_version == "basic-3":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {
                    "category": 3,
                    "spotWidth": spot_width,
                    "spotHeight": spot_height,
                },
            }
        elif self.spot_cleaning_version == "minimal-2":
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {"category": 3, "modifier": modifier, "navigationMode": 1},
            }
        else:  # self.spot_cleaning_version == 'micro-2'
            json = {
                "reqId": "1",
                "cmd": "startCleaning",
                "params": {"category": 3, "navigationMode": 1},
            }

        return self._message(json, STATE_SCHEMA)

    def pause_cleaning(self):
        return self._message({"reqId": "1", "cmd": "pauseCleaning"}, STATE_SCHEMA)

    def resume_cleaning(self):
        return self._message({"reqId": "1", "cmd": "resumeCleaning"}, STATE_SCHEMA)

    def stop_cleaning(self):
        return self._message({"reqId": "1", "cmd": "stopCleaning"}, STATE_SCHEMA)

    def send_to_base(self):
        return self._message({"reqId": "1", "cmd": "sendToBase"}, STATE_SCHEMA)

    def get_robot_state(self):
        return self._message({"reqId": "1", "cmd": "getRobotState"}, STATE_SCHEMA)

    def enable_schedule(self):
        return self._message({"reqId": "1", "cmd": "enableSchedule"}, STANDARD_SCHEMA)

    def disable_schedule(self):
        return self._message({"reqId": "1", "cmd": "disableSchedule"}, STANDARD_SCHEMA)

    def get_schedule(self):
        return self._message({"reqId": "1", "cmd": "getSchedule"}, STANDARD_SCHEMA)

    def locate(self):
        return self._message({"reqId": "1", "cmd": "findMe"}, STANDARD_SCHEMA)

    def get_general_info(self):
        return self._message({"reqId": "1", "cmd": "getGeneralInfo"}, STANDARD_SCHEMA)

    def get_local_stats(self):
        return self._message({"reqId": "1", "cmd": "getLocalStats"}, STANDARD_SCHEMA)

    def get_preferences(self):
        return self._message({"reqId": "1", "cmd": "getPreferences"}, STANDARD_SCHEMA)

    def get_map_boundaries(self, map_id=None):
        return self._message(
            {"reqId": "1", "cmd": "getMapBoundaries", "params": {"mapId": map_id}},
            STANDARD_SCHEMA,
        )

    def get_robot_info(self):
        return self._message({"reqId": "1", "cmd": "getRobotInfo"}, STANDARD_SCHEMA)

    def dismiss_current_alert(self):
        return self._message(
            {"reqId": "1", "cmd": "dismissCurrentAlert"}, STANDARD_SCHEMA
        )

    @property
    def schedule_enabled(self):
        return self.get_robot_state().json()["details"]["isScheduleEnabled"]

    @schedule_enabled.setter
    def schedule_enabled(self, enable):
        if enable:
            self.enable_schedule()
        else:
            self.disable_schedule()

    @property
    def state(self):
        return self.get_robot_state().json()

    @property
    def available_services(self):
        return self.state["availableServices"]

    @property
    def service_version(self):
        return self.available_services["houseCleaning"]

    @property
    def spot_cleaning_version(self):
        return self.available_services["spotCleaning"]


class Auth(requests.auth.AuthBase):
    """Create headers for request authentication"""

    def __init__(self, serial, secret):
        self.serial = serial
        self.secret = secret

    def __call__(self, request):
        # We have to format the date according to RFC 2616
        # https://tools.ietf.org/html/rfc2616#section-14.18

        now = datetime.now(timezone.utc)
        date = format_datetime(now, True)

        try:
            # Attempt to decode request.body (assume bytes received)
            msg = "\n".join([self.serial.lower(), date, request.body.decode("utf8")])
        except AttributeError:
            # Decode failed, assume request.body is already type str
            msg = "\n".join([self.serial.lower(), date, request.body])

        signing = hmac.new(
            key=self.secret.encode("utf8"),
            msg=msg.encode("utf8"),
            digestmod=hashlib.sha256,
        )

        request.headers["Date"] = date
        request.headers["Authorization"] = "NEATOAPP " + signing.hexdigest()

        return request



class NeatoException(Exception):
    """
    General neato exception.
    """


class NeatoLoginException(NeatoException):
    """
    To indicate there is a login issue.
    """


class NeatoRobotException(NeatoException):
    """
    To be thrown anytime there is a robot error.
    """


class NeatoUnsupportedDevice(NeatoRobotException):
    """
    To be thrown only for unsupported devices.
    """



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="pybotvac",
    version="0.0.20",
    description="Python package for controlling Neato pybotvac Connected vacuum robot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Stian Askeland",
    author_email="stianaske@gmail.com",
    url="https://github.com/stianaske/pybotvac",
    license="Licensed under the MIT license. See LICENSE file for details",
    packages=["pybotvac"],
    package_dir={"pybotvac": "pybotvac"},
    package_data={"pybotvac": ["cert/*.crt"]},
    install_requires=["requests", "requests_oauthlib", "voluptuous"],
)



from .neato import Vendor


class Vorwerk(Vendor):
    name = "vorwerk"
    endpoint = "https://beehive.ksecosys.com/"
    passwordless_endpoint = "https://mykobold.eu.auth0.com/passwordless/start"
    token_endpoint = "https://mykobold.eu.auth0.com/oauth/token"  # nosec
    scope = ["openid", "email", "profile", "read:current_user", "offline_access"]
    audience = "https://mykobold.eu.auth0.com/userinfo"
    source = "vorwerk_auth0"



import os
from dataclasses import dataclass
from typing import List, Union


@dataclass(init=False, frozen=True)
class Vendor:
    name: str
    endpoint: str
    auth_endpoint: str
    passwordless_endpoint: str
    token_endpoint: str
    scope: List[str]
    audience: str
    source: str
    cert_path: Union[str, bool] = False
    beehive_version: str = "application/vnd.neato.beehive.v1+json"
    nucleo_version: str = "application/vnd.neato.nucleo.v1"


class Neato(Vendor):
    name = "neato"
    endpoint = "https://beehive.neatocloud.com/"
    auth_endpoint = "https://apps.neatorobotics.com/oauth2/authorize"
    token_endpoint = "https://beehive.neatocloud.com/oauth2/token"  # nosec
    scope = ["public_profile", "control_robots", "maps"]
    cert_path = os.path.join(os.path.dirname(__file__), "cert", "neatocloud.com.crt")



"""Sessionhandling for beehive endpoint."""
import binascii
import json
import os
import os.path
from typing import Callable, Dict, Optional

import requests
from oauthlib.oauth2 import TokenExpiredError
from requests_oauthlib import OAuth2Session

from .exceptions import NeatoException, NeatoLoginException, NeatoRobotException
from .neato import Neato, Vendor
from .vorwerk import Vorwerk

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


class Session:
    def __init__(self, vendor: Vendor):
        """Initialize the session."""
        self.vendor = vendor
        self.endpoint = vendor.endpoint
        self.headers = {"Accept": vendor.beehive_version}

    def get(self, path, **kwargs):
        """Send a GET request to the specified path."""
        raise NotImplementedError

    def urljoin(self, path):
        return urljoin(self.endpoint, path)

    def generate_headers(
        self, custom_headers: Optional[Dict[str, str]] = None
    ) -> Dict[str, str]:
        """"Merge self.headers with custom headers id necessary."""
        if not custom_headers:
            return self.headers

        return {**self.headers, **custom_headers}


class PasswordSession(Session):
    def __init__(self, email: str, password: str, vendor: Vendor = Neato()):
        super().__init__(vendor=vendor)
        self._login(email, password)

    def _login(self, email: str, password: str):
        """
        Login to pybotvac account using provided email and password.

        :param email: email for pybotvac account
        :param password: Password for pybotvac account
        :return:
        """

        try:
            response = requests.post(
                urljoin(self.endpoint, "sessions"),
                json={
                    "email": email,
                    "password": password,
                    "platform": "ios",
                    "token": binascii.hexlify(os.urandom(64)).decode("utf8"),
                },
                headers=self.headers,
            )

            response.raise_for_status()
            access_token = response.json()["access_token"]

            self.headers["Authorization"] = "Token token=%s" % access_token
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            if (
                isinstance(ex, requests.exceptions.HTTPError)
                and ex.response.status_code == 403
            ):
                raise NeatoLoginException(
                    "Unable to login to neato, check account credentials."
                ) from ex
            raise NeatoRobotException("Unable to connect to Neato API.") from ex

    def get(self, path, **kwargs):
        url = self.urljoin(path)
        headers = self.generate_headers(kwargs.pop("headers", None))
        try:
            response = requests.get(url, headers=headers, **kwargs)
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            raise NeatoException("Unable to connect to neato the neato serves.") from ex
        return response


class OAuthSession(Session):
    def __init__(
        self,
        token: Optional[Dict[str, str]] = None,
        client_id: str = None,
        client_secret: str = None,
        redirect_uri: str = None,
        token_updater: Optional[Callable[[str], None]] = None,
        vendor: Vendor = Neato(),
    ):
        super().__init__(vendor=vendor)

        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._token_updater = token_updater

        extra = {"client_id": self._client_id, "client_secret": self._client_secret}

        self._oauth = OAuth2Session(
            auto_refresh_kwargs=extra,
            client_id=client_id,
            token=token,
            redirect_uri=redirect_uri,
            token_updater=token_updater,
            scope=vendor.scope,
        )

    def refresh_tokens(self) -> dict:
        """Refresh and return new tokens."""
        token = self._oauth.refresh_token(f"{self.endpoint}/auth/token")

        if self._token_updater is not None:
            self._token_updater(token)

        return token

    def get_authorization_url(self) -> str:
        """Get an authorization url via oauth2."""
        # pylint: disable=unused-variable
        authorization_url, state = self._oauth.authorization_url(
            self.vendor.auth_endpoint
        )
        return authorization_url

    def fetch_token(self, authorization_response: str) -> Dict[str, str]:
        """Fetch an access token via oauth2."""
        token = self._oauth.fetch_token(
            self.vendor.token_endpoint,
            authorization_response=authorization_response,
            client_secret=self._client_secret,
        )
        return token

    def get(self, path: str, **kwargs) -> requests.Response:
        """Make a get request.

        We don't use the built-in token refresh mechanism of OAuth2 session because
        we want to allow overriding the token refresh logic.
        """
        url = self.urljoin(path)
        try:
            response = self._get(url, **kwargs)
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            raise NeatoException("Unable to connect to neato the neato serves.") from ex
        return response

    def _get(self, path: str, **kwargs) -> requests.Response:
        """Get request without error handling.

        Refreshes the token if necessary.
        """
        headers = self.generate_headers(kwargs.pop("headers", None))
        try:
            return self._oauth.get(path, headers=headers, **kwargs)
        except TokenExpiredError:
            self._oauth.token = self.refresh_tokens()

            return self._oauth.get(path, headers=self.headers, **kwargs)


class PasswordlessSession(Session):
    def __init__(
        self,
        token: Optional[Dict[str, str]] = None,
        client_id: str = None,
        token_updater: Optional[Callable[[str], None]] = None,
        vendor: Vendor = Vorwerk(),
    ):
        super().__init__(vendor=vendor)

        self._token = token
        self._client_id = client_id
        self._token_updater = token_updater

    def send_email_otp(self, email: str):
        """Request an authorization code via email."""
        response = requests.post(
            self.vendor.passwordless_endpoint,
            data=json.dumps(
                {
                    "client_id": self._client_id,
                    "connection": "email",
                    "email": email,
                    "send": "code",
                }
            ),
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()

    def fetch_token_passwordless(self, email: str, code: str):
        """Fetch an access token using the emailed code."""
        response = requests.post(
            self.vendor.token_endpoint,
            data=json.dumps(
                {
                    "prompt": "login",
                    "grant_type": "http://auth0.com/oauth/grant-type/passwordless/otp",
                    "scope": " ".join(self.vendor.scope),
                    "locale": "en",
                    "otp": code,
                    "source": self.vendor.source,
                    "platform": "ios",
                    "audience": self.vendor.audience,
                    "username": email,
                    "client_id": self._client_id,
                    "realm": "email",
                    "country_code": "DE",
                }
            ),
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        self._token = response.json()

    def get(self, path: str, **kwargs) -> requests.Response:
        """Make a get request."""
        url = self.urljoin(path)
        headers = self.generate_headers(kwargs.pop("headers", None))
        headers["Authorization"] = "Auth0Bearer {}".format(self._token.get("id_token"))

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            raise NeatoException("Unable to connect to neato the neato serves.") from ex
        return response
